/*
 * Copyright 2015 - 2022 gary@drinkingtea.net
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/.
 */

#if !defined(OX_USE_STDLIB)

#include "math.hpp"
#include "stringview.hpp"
#include "types.hpp"

#define REG_DEBUG_ENABLE *reinterpret_cast<volatile uint16_t*>(0x4FFF780)
#define REG_DEBUG_FLAGS *reinterpret_cast<volatile uint16_t*>(0x4FFF700)
#define REG_DEBUG_STRING (reinterpret_cast<char*>(0x4FFF600))

namespace ox::hw {
static void (*infoLog)(ox::CRStringView) = [](ox::CRStringView) {};
static void (*debugLog)(ox::CRStringView) = [](ox::CRStringView) {};
static void (*errorLog)(ox::CRStringView) = [](ox::CRStringView) {};
}

namespace mgba {

enum LogChan {
	LOG_FATAL = 0,
	LOG_ERROR = 1,
	LOG_WARN = 2,
	LOG_INFO = 3,
	LOG_DEBUG = 4
};

template<LogChan chan>
static auto mkLogger() {
	return [](ox::CRStringView str) {
		const auto sz = ox::min<std::size_t>(0x100, str.bytes());
		ox_strncpy(REG_DEBUG_STRING, str.data(), sz);
		REG_DEBUG_FLAGS = chan | 0x100;
	};
}

void initConsole() {
	REG_DEBUG_ENABLE = 0xC0DE;
	if (REG_DEBUG_ENABLE == 0x1DEA) {
		ox::hw::infoLog = mgba::mkLogger<LOG_INFO>();
		ox::hw::debugLog = mgba::mkLogger<LOG_INFO>(); // use INFO because mGBA disables DEBUG on start
		ox::hw::errorLog = mgba::mkLogger<LOG_ERROR>();
	}
}

}

#else
#include <iomanip>
#include <iostream>

static const auto OxPrintTrace = std::getenv("OXTRACE") != nullptr;
#endif

#include "strops.hpp"

extern "C" {

void oxTraceInitHook() {
}

void oxTraceHook([[maybe_unused]] const char *file, [[maybe_unused]] int line,
                 [[maybe_unused]] const char *ch, [[maybe_unused]] const char *msg) {
#if defined(OX_USE_STDLIB)
	if (OxPrintTrace) {
		std::cout << std::setw(53) << std::left << ch << "| ";
		std::cout << std::setw(65) << std::left << msg << '|';
		std::cout << " " << file << ':' << line << "\n";
	} else if (ox_strcmp(ch, "debug") == 0 || ox_strcmp(ch, "info") == 0) {
		std::cout << msg << '\n';
	} else if (ox_strcmp(ch, "stdout") == 0) {
		std::cout << msg;
	} else if (ox_strcmp(ch, "stderr") == 0) {
		std::cerr << msg;
	} else if (ox_strcmp(ch, "error") == 0) {
		std::cerr << "\033[31;1;1mERROR:\033[0m (" << file << ':' << line << "): " << msg << '\n';
	}
#else
	if (ox_strcmp(ch, "info") == 0) {
		ox::hw::infoLog(msg);
	} else if (ox_strcmp(ch, "debug") == 0) {
		ox::hw::debugLog(msg);
	} else if (ox_strcmp(ch, "stdout") == 0) {
		ox::hw::infoLog(msg);
	} else if (ox_strcmp(ch, "stderr") == 0) {
		ox::hw::errorLog(msg);
	} else if (ox_strcmp(ch, "error") == 0) {
		ox::hw::errorLog(msg);
	}
#endif
}

}

